# -*- coding: utf-8 -*-
"""
Created on 7 dec. 2018, 16:54
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.utils.misc import compute_next_permutation, strong_connected_components
from svvamp.preferences.profile import Profile


class RuleKemeny(Rule):
    """Kemeny method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleKemeny()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [2 0 1]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [2 1 0]
        w = 0
        score_w = 2
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (True)  ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (True)      ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = lazy, um_option = lazy, icm_option = exact, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 2. 3.]

    Notes
    -----
    We find the order on candidates whose total Kendall tau distance to the voters is minimal. The top element of
    this order is declared the winner. In case several orders are optimal, the first one by lexicographic order is
    given. This implies that if several winners are possible, the one with lowest index is declared the winner.

    For this voting system, even deciding the sincere winner is NP-hard.

    * :meth:`is_cm_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_icm_`: Exact in polynomial time (once the sincere winner is computed).
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia_`: Exact in polynomial time (once the sincere winner is computed).
    * :meth:`is_tm_`: Exact in the time needed to decide the winner of one election, multiplied by :attr:`n_c`.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Mathematics without numbers', J. G. Kemeny, 1959.

    'A Consistent Extension of Condorcet’s Election Principle', H. P. Young and A. Levenglick, 1978.

    'On the approximability of Dodgson and Young elections', Ioannis Caragiannis et al., 2009.

    'Comparing and aggregating partial orders with Kendall tau distances', Franz J. Brandenburg, Andreas Gleißner
    and Andreas Hofmeier, 2013.
    """

    full_name = 'Kemeny'
    abbreviation = 'Kem'

    options_parameters = Rule.options_parameters.copy()
    options_parameters['icm_option'] = {'allowed': ['exact'], 'default': 'exact'}

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=True,
            log_identity="KEMENY", **kwargs
        )

    @cached_property
    def _strong_connected_components_(self):
        return strong_connected_components(self.profile_.matrix_victories_rk > 0)

    @cached_property
    def score_w_(self):
        """Integer. With our convention, ``scores_w_`` = :attr:`n_c` - 1.
        """
        self.mylog("Compute winner's score", 1)
        return self.profile_.n_c - 1

    @cached_property
    def scores_best_to_worst_(self):
        """1d array of integers. With our convention, ``scores_best_to_worst`` is the vector [:attr:`n_c` - 1,
        :attr:`n_c` - 2, ..., 0].
        """
        self.mylog("Compute scores_best_to_worst", 1)
        return np.array(range(self.profile_.n_c - 1, -1, -1))

    @cached_property
    def _count_first_component_(self):
        """Compute the optimal order for the first strongly connected component of the victory matrix (ties count as
        a connection).
        """
        self.mylog("Count first strongly connected component", 1)
        candidates_by_scores_best_to_worst_first_component = []
        order = np.array(sorted(self._strong_connected_components_[0]))
        size_component = len(order)
        best_score = -1
        best_order = None
        while order is not None:
            self.mylogv("order =", order, 3)
            score = np.sum(self.profile_.matrix_duels_rk[:, order][order, :][np.triu_indices(size_component)])
            self.mylogv("score =", score, 3)
            if score > best_score:
                best_order, best_score = order, score
            order = compute_next_permutation(order, size_component)
        self.mylogv("best_order =", best_order, 3)
        w = best_order[0]
        candidates_by_scores_best_to_worst_first_component.extend(best_order)
        return {
            'candidates_by_scores_best_to_worst_first_component': candidates_by_scores_best_to_worst_first_component,
            'w': w}

    @cached_property
    def w_(self):
        return self._count_first_component_['w']

    @cached_property
    def _candidates_by_scores_best_to_worst_first_component_(self):
        return self._count_first_component_['candidates_by_scores_best_to_worst_first_component']

    @cached_property
    def _count_ballots_(self):
        """Compute the optimal order for all the candidates."""
        self.mylog("Count other strongly connected components", 1)
        candidates_by_scores_best_to_worst = self._candidates_by_scores_best_to_worst_first_component_.copy()
        for component_members in self._strong_connected_components_[1:]:
            order = np.array(sorted(component_members))
            size_component = len(order)
            best_score = -1
            best_order = None
            while order is not None:
                self.mylogv("order =", order, 3)
                score = np.sum(self.profile_.matrix_duels_rk[:, order][order, :][np.triu_indices(size_component)])
                self.mylogv("score =", score, 3)
                if score > best_score:
                    best_order, best_score = order, score
                order = compute_next_permutation(order, size_component)
            self.mylogv("best_order =", best_order, 3)
            candidates_by_scores_best_to_worst.extend(best_order)
        candidates_by_scores_best_to_worst = np.array(candidates_by_scores_best_to_worst)
        scores = self.profile_.n_c - 1 - np.argsort(candidates_by_scores_best_to_worst)
        return {'scores': scores, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def scores_(self):
        """1d array of integers. By convention, scores are integers from 1 to :attr:`n_c`, with :attr:`n_c` for the
        winner and 1 for the last candidate in Kemeny optimal order.
        """
        return self._count_ballots_['scores']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. This is an optimal Kemeny order.

        In case several orders are optimal, the first one by lexicographic order is given. This implies that if
        several winners are possible, the one with lowest index is declared the winner.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        return True



# -*- coding: utf-8 -*-
"""
Created on 11 dec. 2018, 13:39
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleTwoRound(Rule):
    """Two Round System.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleTwoRound()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 0]
         [0 0]
         [1 0]
         [2 2]
         [2 2]]
        scores =
        [[2. 1. 2.]
         [3. 0. 2.]]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [[2. 2. 1.]
         [3. 2. 0.]]
        w = 0
        score_w = [2. 3.]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = False
        log_im: im_option = exact
        candidates_im =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = False
        log_um: um_option = exact
        candidates_um =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = False
        log_cm: cm_option = exact
        candidates_cm =
        [0. 0. 0.]
        necessary_coalition_size_cm =
        [0. 2. 4.]
        sufficient_coalition_size_cm =
        [0. 2. 4.]

    Notes
    -----
    Two rounds are actually held, which means that manipulators can change their vote between the first and second
    round. Hence for :attr:`n_c` = 3, this voting system is equivalent to :class:`RuleExhaustiveBallot` (not
    :class:`RuleIRV`). In case of a tie, candidates with lowest index are privileged.

    * :meth:`is_iia_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_cm_`, :meth:`is_icm_`, :meth:`is_im_`, :meth:`is_tm_`, :meth:`is_um_`: Exact in polynomial time.
    """

    full_name = 'Two Round System'
    abbreviation = 'TR'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'im_option': {'allowed': ['exact'], 'default': 'exact'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['exact'], 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_um=False, precheck_icm=False, precheck_tm=False,
            log_identity="TWO_ROUND", **kwargs
        )

    # %% Counting the ballots

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        # First round
        scores_r = np.copy(self.profile_.plurality_scores_rk)
        c = np.argmax(scores_r)
        selected_one = c
        scores_r[c] = 0
        d = np.argmax(scores_r)
        selected_two = d
        # Second round
        w = c if self.profile_.matrix_victories_rk_ctb[c, d] else d
        return {'w': w, 'selected_one': selected_one, 'selected_two': selected_two}

    @cached_property
    def selected_one_(self):
        """Integer. The candidate with highest score at first round."""
        return self._count_ballots_['selected_one']

    @cached_property
    def selected_two_(self):
        """Integer. The candidate with second highest score at first round."""
        return self._count_ballots_['selected_two']

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def ballots_(self):
        """2d array of integers. ``ballots[v, r]`` is the candidate for which voter ``v`` votes at round ``r``,
        where ``r`` = 0 (first round) or ``r`` = 1 (second round).
        """
        ballots = np.zeros((self.profile_.n_v, 2), dtype=np.int)
        ballots[:, 0] = self.profile_.preferences_rk[:, 0]
        c, d = self.selected_one_, self.selected_two_
        ballots[self.profile_.preferences_borda_rk[:, c] > self.profile_.preferences_borda_rk[:, d], 1] = c
        ballots[self.profile_.preferences_borda_rk[:, d] > self.profile_.preferences_borda_rk[:, c], 1] = d
        return ballots

    @cached_property
    def scores_(self):
        """2d array. ``scores[r, c]`` is the number of voters who vote for candidate ``c`` at round ``r``,
        where ``r`` = 0 (first round) or ``r`` = 1 (second round).
        """
        scores = np.zeros((2, self.profile_.n_c))
        scores[0, :] = self.profile_.plurality_scores_rk
        c, d = self.selected_one_, self.selected_two_
        scores[1, c] = self.profile_.matrix_duels_rk[c, d]
        scores[1, d] = self.profile_.matrix_duels_rk[d, c]
        return scores

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. ``candidates_by_scores_best_to_worst[k]`` is the candidate with ``k``\ :sup:`th`
        best score. Finalists are sorted by their score at second round. Other candidates are sorted by their score
        at first round.
        """
        self.mylog("Count candidates_by_scores_best_to_worst", 1)
        return np.argsort(- np.max(self.scores_, 0))

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_majority_favorite_c_rk_ctb(self):
        return True

    # %% Individual manipulation (IM)

    def _compute_im_(self, mode, c=None):
        """Compute IM: is_im, candidates_im, _voters_im and v_im_for_c.

        For Two Round, since calculation is not so expensive, we compute everything, even if complete_mode = False.

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.is_im_v_(0)
            False
            >>> rule.v_im_for_c_
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.],
                   [1., 0., 0.],
                   [1., 0., 0.]])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [1, 2, 0],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.v_im_for_c_
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.]])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 2, 0],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.v_im_for_c_
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 1.],
                   [0., 0., 1.],
                   [0., 0., 0.]])
        """
        self.mylog("Compute IM", 1)
        self._im_was_computed_with_candidates = True
        self._im_was_computed_with_voters = True
        self._im_was_computed_full = True
        self._v_im_for_c = np.zeros((self.profile_.n_v, self.profile_.n_c))
        self._candidates_im = np.zeros(self.profile_.n_c)
        self._voters_im = np.zeros(self.profile_.n_v)
        self._is_im = False
        # First, let us note ``y`` the opponent of ``w`` at second round (in sincere voting). If the second round is
        # held with ``w`` and ``y``, then IM is impossible. Indeed, if ``v`` prefers ``y`` to ``w``, then she cannot
        # do better than her sincere voting.
        #
        # At first round, when ``v`` removes her sincere vote, a candidate ``z``, either ``w`` or ``y``,
        # has the highest score (indeed, only one of them might lose 1 point). Whatever ``v``'s strategic vote,
        # ``z`` will be selected for the second round (either as first selected or second selected). The only thing
        # ``v`` can do is choose ``z``'s opponent, another candidate called ``c_test``. For that, the best she can do
        #  is vote for this ``c_test``.
        #
        #   * If ``z = w``, then choosing ``c_test = y`` is not interesting, because it leads to a second round
        #     ``(w, y)``, not interesting as we noticed.
        #   * If ``z = y``, then choosing ``c_test = w`` is not interesting for the same reason.
        #
        # So, the only votes that ``v`` should consider for first round are candidates ``c_test`` that are not ``w`` or
        # ``y``. In second round, several things can happen, like a duel between ``y`` and ``c_test`` for example. Both
        # candidates might be interesting for the manipulator.
        #
        # Below, ``c_test`` is not the candidate ``v`` wants to get elected, but the candidate that she wants to push
        # to second round.
        for c_test in range(self.profile_.n_c):
            if c_test == self.selected_one_ or c_test == self.selected_two_:
                continue
            if self.scores_[0, self.selected_two_] - 1 + (self.selected_two_ < c_test) > self.scores_[0, c_test] + 1:
                # c_test cannot go to second round at all
                continue
            for v in range(self.profile_.n_v):
                # First round
                scores_first_temp = np.copy(self.scores_[0, :])
                scores_first_temp[self.ballots_[v, 0]] -= 1
                scores_first_temp[c_test] += 1
                selected_one_m = np.argmax(scores_first_temp)
                scores_first_temp[selected_one_m] = -1
                selected_two_m = np.argmax(scores_first_temp)
                # Second round
                score_one_without_v = self.profile_.matrix_duels_rk[selected_one_m, selected_two_m]
                score_two_without_v = self.profile_.matrix_duels_rk[selected_two_m, selected_one_m]
                if (self.profile_.preferences_borda_rk[v, selected_one_m] >
                        self.profile_.preferences_borda_rk[v, selected_two_m]):
                    score_one_without_v -= 1
                else:
                    score_two_without_v -= 1
                # Conclusions
                if (self.v_wants_to_help_c_[v, selected_one_m]
                        and score_one_without_v + 1 + (selected_one_m < selected_two_m) > score_two_without_v):
                    self._v_im_for_c[v, selected_one_m] = True
                    self._candidates_im[selected_one_m] = True
                    self._voters_im[v] = True
                    self._is_im = True
                    continue
                if (self.v_wants_to_help_c_[v, selected_two_m]
                        and score_two_without_v + 1 + (selected_two_m < selected_one_m) > score_one_without_v):
                    self._v_im_for_c[v, selected_two_m] = True
                    self._candidates_im[selected_two_m] = True
                    self._voters_im[v] = True
                    self._is_im = True

    def _compute_im_v_(self, v, c_is_wanted, stop_if_true):
        self._compute_im_(mode='', c=None)

    # %% Trivial Manipulation (TM)

    # Use the general methods.

    # %% Unison manipulation (UM)

    def _um_main_work_c_(self, c):
        """
            >>> profile = Profile([
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_um_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 3, 2],
            ...     [0, 2, 1, 3],
            ...     [1, 2, 0, 3],
            ...     [2, 1, 0, 3],
            ...     [3, 1, 2, 0],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_um_
            array([0., 0., 1., 0.])

            Cases where ``c`` is ``selected_one_s``:

            >>> profile = Profile(preferences_ut=[
            ...     [ 0.5,  0.5, -1. , -0.5],
            ...     [ 0.5, -1. , -1. , -0.5],
            ...     [-0.5,  0.5, -1. , -0.5],
            ...     [-0.5,  0.5,  1. , -0.5],
            ...     [ 0. ,  1. ,  0. ,  1. ],
            ... ], preferences_rk=[
            ...     [0, 1, 3, 2],
            ...     [0, 3, 1, 2],
            ...     [1, 0, 3, 2],
            ...     [2, 1, 3, 0],
            ...     [3, 1, 0, 2],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_um_
            array([1., 0., 0., 1.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 1. ,  0. , -1. ,  1. ,  1. ],
            ...     [ 0. , -0.5,  1. , -0.5,  1. ],
            ...     [ 1. ,  1. ,  0.5,  1. ,  1. ],
            ...     [-0.5, -1. ,  0. ,  1. , -1. ],
            ...     [ 0. , -1. ,  0. ,  0. , -1. ],
            ...     [ 0.5, -0.5,  1. ,  0. ,  1. ],
            ... ], preferences_rk=[
            ...     [0, 4, 3, 1, 2],
            ...     [2, 4, 0, 1, 3],
            ...     [3, 0, 4, 1, 2],
            ...     [3, 2, 0, 4, 1],
            ...     [3, 2, 0, 4, 1],
            ...     [4, 2, 0, 3, 1],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_um_
            array([0., 0., 1., 0., 0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0. ,  0. ,  1. , -0.5, -0.5,  0. , -0.5],
            ...     [ 0. , -0.5, -1. ,  0. , -0.5, -0.5, -0.5],
            ...     [ 0. , -0.5, -0.5, -1. ,  1. , -1. , -0.5],
            ...     [ 0. ,  0. , -1. ,  0.5,  1. ,  0. ,  1. ],
            ...     [ 1. , -1. , -1. , -1. ,  1. ,  1. ,  1. ],
            ...     [-0.5, -1. ,  0.5,  1. ,  0.5, -1. ,  1. ],
            ...     [-1. ,  0. , -1. ,  0.5,  0.5,  0.5,  1. ],
            ... ], preferences_rk=[
            ...     [2, 0, 1, 5, 3, 6, 4],
            ...     [3, 0, 1, 6, 4, 5, 2],
            ...     [4, 0, 6, 2, 1, 3, 5],
            ...     [4, 6, 3, 5, 1, 0, 2],
            ...     [4, 6, 5, 0, 2, 1, 3],
            ...     [6, 3, 4, 2, 0, 1, 5],
            ...     [6, 5, 4, 3, 1, 2, 0],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_um_
            array([0., 0., 0., 0., 1., 0., 0.])
        """
        n_m = self.profile_.matrix_duels_ut.astype(int)[c, self.w_]
        n_s = self.profile_.n_v - n_m
        ballots_first_round_s = self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), 0]
        scores_first_round_s = np.bincount(ballots_first_round_s, minlength=self.profile_.n_c)
        scores_temp = np.copy(scores_first_round_s)
        selected_one_s = np.argmax(scores_temp)
        scores_temp[selected_one_s] = -1
        selected_two_s = np.argmax(scores_temp)
        if c != selected_one_s:
            # Then ``c`` is either ``selected_two_s`` or another candidate. After unison manipulation,
            # ``selected_one_s`` will go to second round anyway (either as first or second selected). So the only
            # interesting strategy at first round is to vote for ``c``. In that case, UM is equivalent to TM.
            # N.B.: the following 'if' is also correct when ``selected_two_s = c``.
            if scores_first_round_s[c] + n_m < scores_first_round_s[selected_two_s] + (selected_two_s < c):
                # c cannot go to second round
                self._candidates_um[c] = False
                return
            # Second round
            d = selected_one_s
            d_vs_c = np.sum(
                self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), d]
                > self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), c])
            c_vs_d = n_s - d_vs_c
            if c_vs_d + n_m < d_vs_c + (d < c):
                self._candidates_um[c] = False
                return
            else:
                self._candidates_um[c] = True
                return
        # Now we know that ``c`` is ``selected_one_s``. So, manipulators can try to choose ``c``'s opponent for
        # second round. ``d`` will be any possible opponent of ``c`` during the second round.
        for d in range(self.profile_.n_c):
            # First round
            # N.B.: the following 'if' is also correct when ``selected_two_s = d``.
            if scores_first_round_s[d] + n_m < scores_first_round_s[selected_two_s] + (selected_two_s < c):
                # d cannot go to second round
                continue
            # Second round
            d_vs_c = np.sum(
                self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), d]
                > self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), c])
            c_vs_d = n_s - d_vs_c
            if c_vs_d + n_m < d_vs_c + (d < c):
                self._candidates_um[c] = False
                return
            else:
                self._candidates_um[c] = True
                return

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods.

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 2, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.candidates_cm_
            array([0., 1., 0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0.5,  0.5],
            ...     [ 1. ,  0.5],
            ...     [ 0. ,  1. ],
            ...     [-0.5,  0. ],
            ...     [-0.5, -0.5],
            ...     [-1. , -1. ],
            ... ], preferences_rk=[
            ...     [0, 1],
            ...     [0, 1],
            ...     [1, 0],
            ...     [1, 0],
            ...     [1, 0],
            ...     [1, 0],
            ... ])
            >>> rule = RuleTwoRound()(profile)
            >>> rule.is_cm_c_with_bounds_(0)
            (False, 3.0, 3.0)
        """
        n_s = self.profile_.n_v - self.profile_.matrix_duels_ut.astype(int)[c, self.w_]
        ballots_first_round_s = self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), 0]
        scores_first_round_s = np.bincount(ballots_first_round_s, minlength=self.profile_.n_c)
        # ``d`` will be any possible opponent of ``c`` during the second round
        for d in range(self.profile_.n_c):
            if d == c:
                continue
            # First round.
            if self.profile_.n_c == 2:
                n_m_first = 0
            else:
                # Besides ``d`` and ``c``, which candidate has the best score?
                scores_temp = np.copy(scores_first_round_s)
                scores_temp[d] = -1
                scores_temp[c] = -1
                e = np.argmax(scores_temp)
                n_m_first = (max(0, np.add(scores_first_round_s[e] - scores_first_round_s[c], e < c))
                             + max(0, np.add(scores_first_round_s[e] - scores_first_round_s[d], e < d)))
            # Second round.
            d_vs_c = np.sum(
                self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), d]
                > self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), c])
            c_vs_d = n_s - d_vs_c
            n_m_second = max(0, d_vs_c - c_vs_d + (d < c))
            # Conclude: how many manipulators are needed?
            self._update_sufficient(self._sufficient_coalition_size_cm, c, max(n_m_first, n_m_second),
                                    'CM: Update sufficient_coalition_size_cm[c] =')
            if self._sufficient_coalition_size_cm[c] == self._necessary_coalition_size_cm[c]:
                return
        self._necessary_coalition_size_cm[c] = self._sufficient_coalition_size_cm[c]



# -*- coding: utf-8 -*-
"""
Created on oct. 20, 2014, 11:03
Copyright François Durand 2014
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import numbers


def is_bool(value):
    """Test whether it is a Boolean.

    Parameters
    ----------
    value : obj
        Value to be tested.

    Returns
    -------
    bool
        True iff ``value`` is a Boolean.

    Examples
    --------
        >>> is_bool(False)
        True
        >>> is_bool(1)
        False
    """
    return isinstance(value, bool)


def is_number(value):
    """Test whether it is a Boolean.

    Parameters
    ----------
    value : obj
        Value to be tested.

    Returns
    -------
    bool
        True iff ``value`` is a Boolean.

    Examples
    --------
        >>> is_number(3)
        True
        >>> is_number(3.)
        True
        >>> is_number('3')
        False
    """
    return isinstance(value, numbers.Number)



# -*- coding: utf-8 -*-
"""
Created on 16 jul. 2021, 15:49
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleSplitCycle(Rule):
    """Split Cycle.

    Examples
    --------
    Split Cycle does not :attr:`meets_condorcet_c_ut_rel`:

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. ,  0. ,  0. ],
        ...     [ 0.5,  1. , -0.5],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ... ])
        >>> RuleSplitCycle()(profile).w_
        0
        >>> profile.condorcet_winner_ut_rel
        1
    """

    full_name = 'Split Cycle'
    abbreviation = 'SC'

    options_parameters = Rule.options_parameters.copy()
    options_parameters['icm_option'] = {'allowed': ['exact'], 'default': 'exact'}

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="SPLIT_CYCLE", **kwargs
        )

    # %% Counting the ballots

    # noinspection PyProtectedMember
    @staticmethod
    def _count_ballot_aux(matrix_duels):
        """

        Parameters
        ----------
        matrix_duels

        Returns
        -------

        Examples
        --------
            >>> a, b, c, d, e, f = 1, 2, 3, 4, 5, 6
            >>> def cowinners(profile):
            ...     results = RuleSplitCycle._count_ballot_aux(profile.matrix_duels_rk)
            ...     better_or_tie_c = results['better_or_tie_c']
            ...     return list(np.where(better_or_tie_c == np.max(better_or_tie_c))[0])

            >>> profile = Profile(preferences_rk=
            ...     [[a, c, b]] * 5
            ...     + [[c, b, a]] * 4
            ...     + [[b, a, c]] * 4
            ...     + [[c, a, b]] * 2
            ...     + [[b, c, a]] * 2
            ... )
            >>> cowinners(profile)
            [2]

            >>> profile = Profile(preferences_rk=(
            ...     [[d, b, a, c]] * 3
            ...     + [[a, d, c, b]] * 1
            ...     + [[a, c, b, d]] * 2
            ...     + [[c, b, d, a]] * 1
            ...     + [[c, b, a, d]] * 2
            ... ))
            >>> cowinners(profile)
            [0, 1, 2]

            >>> profile = Profile(preferences_rk=(
            ...     [[b,a,c,e,d,f]] * 2
            ...     + [[e,d,f,b,a,c]] * 2
            ...     + [[a,c,b,d,f,e]] * 2
            ...     + [[d,f,e,a,c,b]] * 2
            ...     + [[c,b,a,f,e,d]] * 2
            ...     + [[f,e,d,c,b,a]] * 2
            ...     + [[a,f,b,c,d,e]] * 2
            ...     + [[e,d,c,b,a,f]] * 2
            ...     + [[d,a,b,c,e,f]] * 1
            ...     + [[f,e,c,b,d,a]] * 1
            ... ))
            >>> cowinners(profile)
            [0, 1, 2, 3, 4]

            >>> profile = Profile(preferences_rk=(
            ...     [[e, c, b, a, d]] * 4
            ...     + [[c, b, a, d, e]] * 3
            ...     + [[e, b, a, d, c]] * 7
            ...     + [[d, a, c, b, e]] * 7
            ...     + [[c, d, b, a, e]] * 2
            ...     + [[d, e, a, b, c]] * 1
            ... ))
            >>> cowinners(profile)
            [3]

            >>> profile = Profile(preferences_rk=(
            ...     [[d, b, c, a]] * 1
            ...     + [[c, a, d, b]] * 2
            ...     + [[c, d, b, a]] * 1
            ...     + [[b, a, d, c]] * 2
            ... ))
            >>> cowinners(profile)
            [1, 2, 3]

            >>> profile = Profile(preferences_rk=(
            ...     [[b, a, c, d]] * 9
            ...     + [[a, c, b, d]] * 3
            ...     + [[d, a, c, b]] * 2
            ...     + [[d, c, b, a]] * 5
            ...     + [[d, c, a, b]] * 4
            ... ))
            >>> cowinners(profile)
            [0, 1, 2]

            >>> profile = Profile(preferences_rk=(
            ...     [[b, a, c, d, e]] * 2
            ...     + [[b, a, c, e, d]] * 3
            ...     + [[a, e, c, b, d]] * 3
            ...     + [[d, e, a, c, b]] * 1
            ...     + [[d, a, e, c, b]] * 1
            ...     + [[d, e, c, b, a]] * 5
            ...     + [[b, e, a, c, d]] * 4
            ...     + [[d, c, a, b, e]] * 4
            ... ))
            >>> cowinners(profile)
            [0, 1, 3]

            >>> profile = Profile(preferences_rk=(
            ...     [[a, b, c, d]] * 5
            ...     + [[b, c, d, a]] * 3
            ...     + [[a, c, b, d]] * 1
            ...     + [[d, a, b, c]] * 7
            ...     + [[c, a, b, d]] * 4
            ...     + [[d, b, c, a]] * 4
            ...     + [[b, a, c, d]] * 2
            ...     + [[b, a, d, c]] * 11
            ...     + [[c, d, a, b]] * 16
            ... ))
            >>> cowinners(profile)
            [0]

            >>> profile = Profile(preferences_rk=(
            ...     [[a, b, c, d, e]] * 5
            ...     + [[b, c, e, d, a]] * 3
            ...     + [[c, e, d, a, b]] * 4
            ...     + [[e, a, c, b, d]] * 1
            ...     + [[d, a, b, c, e]] * 7
            ...     + [[c, a, e, b, d]] * 4
            ...     + [[e, d, b, c, a]] * 4
            ...     + [[e, b, a, c, d]] * 2
            ...     + [[e, b, a, d, c]] * 7
            ...     + [[b, e, a, d, c]] * 4
            ...     + [[c, d, a, e, b]] * 12
            ... ))
            >>> cowinners(profile)
            [0, 2]

            >>> profile = Profile(preferences_rk=(
            ...     [[c, b, a]] * 3
            ...     + [[b, a, c]] * 3
            ...     + [[a, c, b]] * 3
            ... ))
            >>> cowinners(profile)
            [0, 1, 2]

            >>> profile = Profile(preferences_rk=(
            ...     [[c, b, a, d]] * 3
            ...     + [[b, a, d, c]] * 2
            ...     + [[d, b, a, c]] * 1
            ...     + [[d, a, c, b]] * 3
            ... ))
            >>> cowinners(profile)
            [0, 1, 3]
        """
        n_c = matrix_duels.shape[0]
        # Margin graph
        margin_graph = np.array(matrix_duels - matrix_duels.T, dtype=float)
        margin_graph[margin_graph <= 0] = -np.inf
        # Weak Condorcet winners
        weak_condorcet_winners = np.all(margin_graph >= 0, axis=1)
        # Calculate widest path for candidate c to d
        strength = margin_graph.copy()
        for i in range(n_c):
            for j in range(n_c):
                if i == j:
                    continue
                if not weak_condorcet_winners[j]:
                    for k in range(n_c):
                        if k == i or k == j:
                            continue
                        strength[j][k] = max(strength[j][k], min(strength[j][i], strength[i][k]))
        # Winners
        better_or_tie_c = np.sum(strength >= margin_graph.T, 1)
        candidates_by_scores_best_to_worst = np.argsort(- better_or_tie_c, kind='mergesort')
        winners = np.where(better_or_tie_c == np.max(better_or_tie_c))[0]
        w = candidates_by_scores_best_to_worst[0]
        return {'w': w, 'winners': winners, 'margin_graph': margin_graph, 'strength': strength,
                'better_or_tie_c': better_or_tie_c}

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        return self._count_ballot_aux(self.profile_.matrix_duels_rk)

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def scores_(self):
        return self._count_ballots_['better_or_tie_c']

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        # TO DO : to check
        return True



# -*- coding: utf-8 -*-
"""
Created on 13 jul. 2021, 06:22
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.rules.rule_irv import RuleIRV
from svvamp.preferences.profile import Profile
from svvamp.utils.util_cache import cached_property
from svvamp.utils.pseudo_bool import equal_true
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut, matrix_victories_to_smith_set


class RuleTideman(Rule):
    """Tideman rule (combination of Smith set and IRV-like elimination).

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleTideman()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[1 0 0]]
        candidates_by_scores_best_to_worst
        [0 1 2]
        scores_best_to_worst
        [[1 0 0]]
        w = 0
        score_w = [1]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False

    Tideman does not :attr:`meets_condorcet_c_ut_abs_ctb`:

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. ,  1. , -0.5],
        ...     [ 0.5,  0. ,  0.5],
        ... ], preferences_rk=[
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ... ])
        >>> RuleTideman()(profile).w_
        1
        >>> profile.condorcet_winner_ut_abs_ctb
        0

    Tideman does not :attr:`meets_condorcet_c_ut_rel`:

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. ,  1. , -1. ],
        ...     [ 0.5,  1. , -0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ... ])
        >>> RuleTideman()(profile).w_
        0
        >>> profile.condorcet_winner_ut_rel
        1

    Notes
    -----
    Each voter must provide a strict total order. Restrict the election to the Smith Set (in the sense of
    :attr:`smith_set_rk`), then eliminate the plurality loser, then iterate.

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Rely on :class:`RuleIRV`'s fast algorithm. Polynomial heuristic. Can prove
          CM but unable to decide non-CM (except in rare obvious cases).
        * :attr:`cm_option` = ``'slow'``: Rely on :class:`RuleExhaustiveBallot`'s exact algorithm. Non-polynomial
          heuristic (:math:`2^{n_c}`). Quite efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'very_slow'``: Rely on :class:`RuleIRV`'s exact algorithm. Non-polynomial
          heuristic (:math:`n_c!`). Very efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

        Each algorithm above exploits the faster ones. For example, if :attr:`cm_option` = ``'very_slow'``,
        SVVAMP tries the fast algorithm first, then the slow one, then the 'very slow' one. As soon as it reaches
        a decision, computation stops.

    References
    ----------
    'Four Condorcet-Hare Hybrid Methods for Single-Winner Elections', James Green-Armytage, 2011.
    """

    full_name = 'Tideman'
    abbreviation = 'Tid'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'cm_option': {'allowed': {'fast', 'slow', 'very_slow', 'exact'}, 'default': 'fast'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="TIDEMAN", **kwargs
        )

    def __call__(self, profile):
        self.delete_cache(suffix='_')
        self.profile_ = profile
        # Grab the IRV ballot of the profile (or create it)
        irv_options = {}
        if self.cm_option == 'fast':
            irv_options['cm_option'] = 'fast'
        elif self.cm_option == 'slow':
            irv_options['cm_option'] = 'slow'
        else:  # self.cm_option in {'very_slow', 'exact'}:
            irv_options['cm_option'] = 'exact'
        self.irv_ = RuleIRV(**irv_options)(self.profile_)
        return self

    # %% Counting the ballots

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        scores = []
        candidates_alive = np.array(range(self.profile_.n_c))
        elimination_path = []
        while True:
            # Eliminate candidates outside Smith set
            # Here, scores_r is 1 for being in the Smith set, 0 otherwise
            matrix_victories_rk_r = self.profile_.matrix_victories_rk[candidates_alive, :][:, candidates_alive]
            smith_set_r = matrix_victories_to_smith_set(matrix_victories_rk_r)
            # La suite
            candidates_alive_new = candidates_alive[smith_set_r]
            scores_r = [(1 if c in candidates_alive_new else 0) for c in range(self.profile_.n_c)]
            scores.append(scores_r)
            eliminated = sorted(set(candidates_alive) - set(candidates_alive_new), reverse=True)
            elimination_path.extend(eliminated)
            candidates_alive = candidates_alive_new
            if len(candidates_alive) == 1:
                break
            # Eliminate candidate with the lowest Plurality score
            is_candidate_alive = np.zeros(self.profile_.n_c, dtype=bool)
            is_candidate_alive[candidates_alive] = True
            scores_r = np.full(self.profile_.n_c, np.nan)
            scores_r[is_candidate_alive] = np.sum(
                self.profile_.preferences_borda_rk[:, is_candidate_alive]
                == np.max(self.profile_.preferences_borda_rk[:, is_candidate_alive], 1)[:, np.newaxis],
                0)
            scores.append(scores_r)
            loser = np.where(scores_r == np.nanmin(scores_r))[0][-1]
            self.mylogv("loser =", loser, 3)
            candidates_alive = np.array([c for c in candidates_alive if c != loser])
            elimination_path.append(loser)
            if len(candidates_alive) == 1:
                break
        w = candidates_alive[0]
        elimination_path.append(w)
        return {'w': w, 'scores': np.array(scores), 'elimination_path': np.array(elimination_path)}

    @cached_property
    def w_(self):
        self.mylog("Compute w", 1)
        plurality_elimination_engine = self.profile_.plurality_elimination_engine()
        plurality_loser = -1
        while True:
            # Eliminate candidates outside Smith set
            candidates_alive = plurality_elimination_engine.candidates_alive
            matrix_victories_rk_r = self.profile_.matrix_victories_rk[candidates_alive, :][:, candidates_alive]
            smith_set_r = matrix_victories_to_smith_set(matrix_victories_rk_r)
            if len(smith_set_r) == 1:
                return candidates_alive[smith_set_r][0]
            smith_losers = np.delete(candidates_alive, smith_set_r)
            # Eliminate candidate with the lowest Plurality score
            if plurality_loser != -1:
                plurality_elimination_engine.update_scores()
            for loser in smith_losers:
                plurality_elimination_engine.eliminate_candidate_and_update_scores(loser)
            scores = plurality_elimination_engine.scores
            if plurality_elimination_engine.nb_candidates_alive == 2:
                return np.nanargmax(scores)
            plurality_loser = np.where(scores == np.nanmin(scores))[0][-1]  # Tie-breaking: the last index
            plurality_elimination_engine.eliminate_candidate(plurality_loser)

    @cached_property
    def scores_(self):
        return self._count_ballots_['scores']

    @cached_property
    def elimination_path_(self):
        return self._count_ballots_['elimination_path']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        return np.array(self.elimination_path_[::-1])

    # TO DO: implement v_might_im_for_c_

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_majority_favorite_c_rk_ctb(self):
        return True

    @cached_property
    def meets_condorcet_c_rk(self):
        return True

    # %% Individual manipulation (IM)

    # TODO: should be implemented.

    # %% Trivial Manipulation (TM)

    # Use the general methods from class Rule.

    # %% Unison manipulation (UM)

    def _um_preliminary_checks_c_(self, c):
        if self.um_option not in {'fast', 'lazy'} or self.cm_option not in {'fast', 'lazy'}:
            if (
                self.w_ == self.profile_.condorcet_winner_rk_ctb
                and not self.profile_.c_might_be_there_when_cw_is_eliminated_irv_style[c]
            ):
                # Impossible to manipulate with n_m manipulators
                self._candidates_um[c] = False

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods from class Rule.

    # %% Coalition Manipulation (CM)

    def _cm_preliminary_checks_c_subclass_(self, c, optimize_bounds):
        """CM: preliminary checks for challenger ``c``.

        Let us consider the case where `w` is the Condorcet winner (rk ctb). Then even after manipulation, `w` beats
        `c`. At any step, we must always have `c` in the (manipulated) Smith Set, hence we must have `w` too.
        As at some point, `w` must be eliminated, IRV-style, at a moment where `c` is still present in the election.
        """
        if self.cm_option != 'fast':
            if (
                self.w_ == self.profile_.condorcet_winner_rk_ctb
                and not self.profile_.c_might_be_there_when_cw_is_eliminated_irv_style[c]
            ):
                # Impossible to manipulate with n_m manipulators
                n_m = self.profile_.matrix_duels_ut[c, self.w_]
                self._update_necessary(self._necessary_coalition_size_cm, c, n_m + 1,
                                       'CM: Update necessary_coalition_size_cm[c] = n_m + 1 =')

    @property
    def losing_candidates_(self):
        """If ``irv_.w_ does not win, then we put her first. Other losers are sorted as usual. (scores in
        ``matrix_duels_ut``).
        """
        self.mylog("Compute ordered list of losing candidates", 1)
        if self.w_ == self.irv_.w_:
            # As usual
            losing_candidates = np.concatenate((
                np.array(range(0, self.w_), dtype=int), np.array(range(self.w_ + 1, self.profile_.n_c), dtype=int)))
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
        else:
            # Put irv_.w_ first.
            losing_candidates = np.array(
                [c for c in range(self.profile_.n_c) if c != self.w_ and c != self.irv_.w_]).astype(np.int)
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
            losing_candidates = np.concatenate(([self.irv_.w_], losing_candidates))
        return losing_candidates

    def _cm_aux_(self, c, ballots_m, preferences_rk_s):
        profile_test = Profile(preferences_rk=np.concatenate((preferences_rk_s, ballots_m)), sort_voters=False)
        if profile_test.n_v != self.profile_.n_v:
            raise AssertionError('Uh-oh!')
        winner_test = self.__class__()(profile_test).w_
        return winner_test == c

    def _cm_main_work_c_(self, c, optimize_bounds):
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        candidates = np.array(range(self.profile_.n_c))
        preferences_borda_s = self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        preferences_rk_s = self.profile_.preferences_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        matrix_duels_vtb_temp = (preferences_ut_to_matrix_duels_ut(preferences_borda_s))
        self.mylogm("CM: matrix_duels_vtb_temp =", matrix_duels_vtb_temp, 3)
        # More preliminary checks. It's more convenient to put them in that method, because we need
        # ``preferences_borda_s`` and ``matrix_duels_vtb_temp``.
        d_neq_c = (np.array(range(self.profile_.n_c)) != c)
        # Prevent another cond. Look at the weakest duel for ``d``, she has ``matrix_duels_vtb_temp[d, e]``. We simply
        # need that:
        # ``matrix_duels_vtb_temp[d, e] <= (n_s + n_m) / 2``
        # ``2 * max_d(min_e(matrix_duels_vtb_temp[d, e])) - n_s <= n_m``
        n_manip_prevent_cond = 0
        for d in candidates[d_neq_c]:
            e_neq_d = (np.array(range(self.profile_.n_c)) != d)
            n_prevent_d = np.maximum(2 * np.min(matrix_duels_vtb_temp[d, e_neq_d]) - n_s, 0)
            n_manip_prevent_cond = max(n_manip_prevent_cond, n_prevent_d)
        self.mylogv("CM: n_manip_prevent_cond =", n_manip_prevent_cond, 3)
        self._update_necessary(self._necessary_coalition_size_cm, c, n_manip_prevent_cond,
                               'CM: Update necessary_coalition_size_cm[c] = n_manip_prevent_cond =')
        if not optimize_bounds and self._necessary_coalition_size_cm[c] > n_m:
            return True

        # Let us work
        if self.w_ == self.irv_.w_:
            self.mylog('CM: c != self.irv_.w == self.w', 3)
            if self.cm_option == "fast":
                self.irv_.cm_option = "fast"
            elif self.cm_option == "slow":
                self.irv_.cm_option = "slow"
            else:
                self.irv_.cm_option = "exact"
            irv_is_cm_c = self.irv_.is_cm_c_(c)
            if equal_true(irv_is_cm_c):
                # Use IRV without bounds
                self.mylog('CM: Use IRV without bounds')
                suggested_path_one = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_one, 3)
                ballots_m = self.irv_.example_ballots_cm_c_(c)
                manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    # We will not do better with any algorithm (even the brute force algo).
                    return False
                # Use IRV with bounds
                self.irv_.is_cm_c_with_bounds_(c)
                self.mylog('CM: Use IRV with bounds')
                suggested_path_two = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_two, 3)
                if np.array_equal(suggested_path_one, suggested_path_two):
                    self.mylog('CM: Same suggested path as before, skip computation')
                else:
                    ballots_m = self.irv_.example_ballots_cm_c_(c)
                    manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                    self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                    if manipulation_found:
                        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                                'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                        # We will not do better with any algorithm (even the brute force algo).
                        return False
        else:  # self.w_ != self.irv_.w_:
            if c == self.irv_.w_:
                self.mylog('CM: c == self.irv_.w != self._w', 3)
                suggested_path = self.irv_.elimination_path_
                self.mylogv("CM: suggested_path =", suggested_path, 3)
                ballots_m = self.irv_.example_ballots_cm_w_against_(w_other_rule=self.w_)
                manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    # We will not do better with any algorithm (even the brute force algo).
                    return
            else:
                self.mylog('CM: c, self.irv_.w_ and self.w_ are all distinct', 3)
        if self.cm_option == 'exact':
            return self._cm_main_work_c_exact_(c, optimize_bounds)
